/**********************************************************************

	FILE: md3imp.cpp

	DESCRIPTION:  Quake III .MD3 file import module

 **********************************************************************/

#include "Max.h"
#include <stdio.h>
#include <direct.h>
#include <commdlg.h>
#include "stdmat.h"
#include "md3imp.h"
#include "md3ires.h"
#include "md3Model.h"

// The file stream


HINSTANCE hInstance;

TCHAR *GetString(int id)
	{
	static TCHAR buf[256];
	if (hInstance)
		return LoadString(hInstance, id, buf, sizeof(buf)) ? buf : NULL;
	return NULL;
	}

// Jaguar interface code

BOOL WINAPI DllMain(HINSTANCE hinstDLL,ULONG fdwReason,LPVOID lpvReserved) {
	hInstance = hinstDLL;
	switch(fdwReason) {
		case DLL_PROCESS_ATTACH:
			//
			break;
		case DLL_THREAD_ATTACH:
			break;
		case DLL_THREAD_DETACH:
			break;
		case DLL_PROCESS_DETACH:
			break;
		}
	return(TRUE);
	}


//------------------------------------------------------

class MD3ClassDesc:public ClassDesc {
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading = FALSE) { return new MD3Import; }
	const TCHAR *	ClassName() { return GetString(IDS_TH_QUAKE3); }
	SClass_ID		SuperClassID() { return SCENE_IMPORT_CLASS_ID; }
	Class_ID		ClassID() { return Class_ID(2002,0715); }
	const TCHAR* 	Category() { return GetString(IDS_TH_SCENEIMPORT);  }
	};

static MD3ClassDesc MD3Desc;

//------------------------------------------------------
// This is the interface to Jaguar:
//------------------------------------------------------

__declspec( dllexport ) const TCHAR *
LibDescription() { return GetString(IDS_TH_MD3IMPORTDLL); }

__declspec( dllexport ) int
LibNumberClasses() { return 1; }

__declspec( dllexport ) ClassDesc *
LibClassDesc(int i) {
	switch(i) {
		case 0: return &MD3Desc; break;
		default: return 0; break;
		}

	}

// Return version so can detect obsolete DLLs
__declspec( dllexport ) ULONG 
LibVersion() { return VERSION_3DSMAX; }

// Let the plug-in register itself for deferred loading
__declspec( dllexport ) ULONG CanAutoDefer()
{
	return 1;
}

//
// .MD3 import module functions follow:
//

MD3Import::MD3Import() {
	}

MD3Import::~MD3Import() {
	}

int
MD3Import::ExtCount() {
	return 1;
	}

const TCHAR *
MD3Import::Ext(int n) {		// Extensions supported for import/export modules
	switch(n) {
		case 0:
			return _T("MD3");
		}
	return _T("");
	}

const TCHAR *
MD3Import::LongDesc() {			// Long ASCII description (i.e. "Targa 2.0 Image File")
	return GetString(IDS_TH_QUAKE3MD3FILE);
	}
	
const TCHAR *
MD3Import::ShortDesc() {			// Short ASCII description (i.e. "Targa")
	return GetString(IDS_TH_QUAKE3);
	}

const TCHAR *
MD3Import::AuthorName() {			// ASCII Author name
	return GetString(IDS_TH_AUTH);
	}

const TCHAR *
MD3Import::CopyrightMessage() {	// ASCII Copyright message
	return GetString(IDS_TH_COPYRIGHT_FREE);
	}

const TCHAR *
MD3Import::OtherMessage1() {		// Other message #1
	return _T("");
	}

const TCHAR *
MD3Import::OtherMessage2() {		// Other message #2
	return _T("");
	}

unsigned int
MD3Import::Version() {				// Version number * 100 (i.e. v3.01 = 301)
	return 101;
	}

void
MD3Import::ShowAbout(HWND hWnd) {			// Optional
 	}

void MeshDeleteFaces(Mesh &mesh,BitArray& set)
	{
	int j = 0;
	for (int i=0; i<mesh.getNumFaces(); i++) {
		if (set[i]) continue;		
		mesh.faces[j] = mesh.faces[i];		
		if (mesh.tvFace) {
			mesh.tvFace[j] = mesh.tvFace[i];
			}
		j++;
		}
	if (j!=mesh.getNumFaces()) {
		mesh.setNumFaces(j,TRUE);
		}
	}

void MeshDeleteVertices(Mesh &mesh,BitArray& set)
	{
	int j = 0;
	for (int i=0; i<mesh.getNumVerts(); i++) {
		if (set[i]) continue;		
		mesh.verts[j] = mesh.verts[i];		
		j++;
		}
	if (j!=mesh.getNumVerts()) {
		mesh.setNumVerts(j,TRUE);
		}
	}


void WeldVertSet(Mesh *mesh,BitArray &set)
	{
	int i;
	int firstSel = -1, numSel=0;
	Point3 pt(0,0,0), dt;

	// Find the first selected vert and find the average of all
	// selected verts
	for (i=0; i<set.GetSize(); i++) {
		if (set[i]) {
			if (firstSel<0) firstSel = i;
			numSel++;
			pt += mesh->verts[i];
			}
		}
	if (firstSel < 0) return;
	pt /= float(numSel);

	// Move the first point to the average position of all verts
	// getting welded.
	dt = pt-mesh->verts[firstSel];
	mesh->verts[firstSel] = pt;

	// Go through each face. If exactly one of the faces vertices is
	// going to be welded, Then remap it. If more than one is going to
	// be welded then delete the face.
	BitArray delSet;
	delSet.SetSize(mesh->getNumFaces());
	for (i=0; i<mesh->getNumFaces(); i++) {
		int ct = 0;
		for (int j=0; j<3; j++) {
			if (set[mesh->faces[i].v[j]]) ct++;			
			}

		if (ct > 1) delSet.Set(i); // Delete it
		else if (ct == 1) {
			// Remap it.
			DWORD v[3] = {mesh->faces[i].v[0],mesh->faces[i].v[1],mesh->faces[i].v[2]};
			for (int j=0; j<3; j++) {
				if (set[mesh->faces[i].v[j]]) {
					v[j] = firstSel;
					mesh->faces[i].v[j] = firstSel;
					break;
					}
				}
			}
		}
	int j=0;
	for (i=0; i<mesh->getNumFaces(); i++) {
		if (!delSet[i])
			j++;
		}
	MeshDeleteFaces(*mesh,delSet);	
	
	// Build the delMap. This is a table with one entry per vertex. The
	// entry is the number of vertices deleted before this vertex.
	DWORDTab delMap;
	DWORD deleted = 0;
	delMap.SetCount(mesh->getNumVerts());
	for (i=0; i<mesh->getNumVerts(); i++) {
		delMap[i] = deleted;
		if (set[i] && i!=firstSel) {
			deleted++;
			}
		}

	// Now remap the faces using the delMap
	for (i=0; i<mesh->getNumFaces(); i++) {
		DWORD v[3];
		v[0] = mesh->faces[i].v[0] - delMap[mesh->faces[i].v[0]];
		v[1] = mesh->faces[i].v[1] - delMap[mesh->faces[i].v[1]];
		v[2] = mesh->faces[i].v[2] - delMap[mesh->faces[i].v[2]];		
		mesh->faces[i].v[0] = v[0];
		mesh->faces[i].v[1] = v[1];
		mesh->faces[i].v[2] = v[2];
		}

	// Go through all selected vertices, except the first and delete them.	
	set.Set(firstSel,FALSE);
	j=firstSel+1;
	for (i=firstSel+1; i<mesh->getNumVerts(); i++) {
		if (!set[i])
			j++;
		}
	MeshDeleteVertices(*mesh,set);

	mesh->InvalidateEdgeList();
	mesh->InvalidateGeomCache();	
	}



static void WeldVerts(Mesh *mesh,float thresh)
	{
	int i, j;	
	// So we can compare length squared
	thresh = thresh * thresh;

#ifdef DBGFINAL
int origVerts = mesh->getNumVerts();
#endif // DBGFINAL
	
	for (i=0; i<mesh->getNumVerts(); i++) {
		// We'll flag verts that are within the threshold for the ith vert
		BitArray set;
		set.SetSize(mesh->getNumVerts());			
		set.Set(i);
		int found = 0;

		for (j=0; j<mesh->getNumVerts(); j++) {
			if (i==j) continue;

			if (LengthSquared(mesh->verts[i]-mesh->verts[j]) < thresh) {
				set.Set(j);
				found++;
				}
			}
		
		// If we found any to weld, weld away!
		if (found)
			WeldVertSet(mesh,set);
		}

#ifdef DBGFINAL
int finalVerts = mesh->getNumVerts();
if(finalVerts != origVerts)
	DebugPrint("Weld dropped vert count to %d from %d\n",finalVerts,origVerts);
#endif // DBGFINAL
	}


char *GetFilename(char *path)
{
	char *p=NULL;
	if(strchr(path,'/')||strchr(path,'\\'))
	{			
		for(p=path+strlen(path);p>path&&(*p!='\\'&&*p!='/');p--){if(*p=='\n')*p='\0';}
		return p+1;
	}
	return path;
}
bool IsFileExist(char *file)
{
	FILE *in = 0;
	if(in = fopen(file, "r"))
	{
		fclose(in);		
		return true;
	}	
	return false;
}
char *FindTexture(char *path,char *mdl,char *obj,char *shd)
{
	static char fullpath[256];
	char *file=NULL;
	if(shd)
	{
		file=GetFilename(shd);
		if(file)
		{			
			sprintf(fullpath,"%s%s",path,file);
			if(IsFileExist(fullpath))return fullpath;
		}
	}	
	sprintf(fullpath,"%s%s_default.skin",path,mdl);
	if(!IsFileExist(fullpath))sprintf(fullpath,"%s%s_red.skin",path,mdl);
	else if(!IsFileExist(fullpath))sprintf(fullpath,"%s%s_blue.skin",path,mdl);
	FILE *in = 0;
	if(!(in=fopen(fullpath, "r")))return NULL;
    char line[1024];
	while (fgets(line, 1024, in))
	{		
		if(strncmp(line,obj,strlen(obj))==0)
		{			
			file=GetFilename(line);
			if(file)
			{				
				sprintf(fullpath,"%s%s",path,file);
				if(IsFileExist(fullpath))return fullpath;
			}
			break;
		}
	}
	fclose(in);
	return NULL;
}

int
MD3Import::DoImport(const TCHAR *filename,ImpInterface *i,Interface *gi, BOOL suppressPrompts) {

	md3Model_t model;
	if (model.Load(filename) == md3ModelErrorNone) {
		char path[256];char modelname[64];
		strncpy(path,filename,255);
		char *p=NULL;
		for(p=path+strlen(path);p>path&&(*p!='\\'&&*p!='/');p--){;}
		strncpy(modelname,p+1,63);
		*(p+1)='\0';
		for(p=modelname+strlen(modelname);p>modelname&&*p!='.';p--){;}
		*p='\0';		
		
		char objectname[256];
		md3Surface_t *surfaces = model.GetSurfaces();
		int numSurf = model.GetNumSurfaces();
		for(int sidx=0;sidx<numSurf;++sidx)
		{
			TriObject *object = CreateNewTriObject();
			if(!object)
				return 0;
			strncpy(objectname,surfaces[sidx].name,255);
			if(!strlen(objectname))sprintf(objectname,"md3_%02d",sidx);
			Mesh *msh=&object->GetMesh();
			StdMat *m = NewDefaultStdMat();
			m->SetName(objectname);
		//	m->SetDiffuse(Color(1.0f,1.0f,1.0f),0);

			char *shader=surfaces[sidx].shaders[0].name;
			char *skin=FindTexture(path,modelname,objectname,shader);
			if(skin)
			{
				BitmapTex *bmt = NewDefaultBitmapTex();
				bmt->SetMapName(skin);
				m->SetSubTexmap(1,bmt);
				m->EnableMap(1,true);
			}
	
			msh->setNumVerts(surfaces[sidx].numVerts);
			msh->setNumFaces(surfaces[sidx].numTriangles);
			msh->setNumTVerts(surfaces[sidx].numVerts);
			msh->setNumTVFaces(surfaces[sidx].numTriangles);
			md3Vertex_t *verts=surfaces[sidx].verts;

			for(int vidx=0;vidx<surfaces[sidx].numVerts;++vidx)
			{

				msh->setVert(vidx, (float *)verts[vidx].pos);
				msh->setTVert(vidx, UVVert(surfaces[sidx].tex[vidx][0],surfaces[sidx].tex[vidx][1],0.0f));
				msh->setNormal(vidx, (float *)verts[vidx].norm);
			}
			md3Triangle_t *tris=surfaces[sidx].tris;

			for(int fidx=0;fidx<surfaces[sidx].numTriangles;++fidx)
			{
				msh->faces[fidx].setVerts(tris[fidx][0],tris[fidx][1],tris[fidx][2]);
				msh->tvFace[fidx].setTVerts(tris[fidx][0],tris[fidx][1],tris[fidx][2]);
				msh->faces[fidx].setEdgeVisFlags(1,1,1);
			}

			WeldVerts(msh,0.001f);

			ImpNode *node = i->CreateNode();
			if(!node) {
				delete object;
				return 0;
			}

			Matrix3 tm;
			tm.IdentityMatrix();
			node->Reference(object);
			node->SetTransform(0,tm);
			i->AddNodeToScene(node);
			if(m)
			{
				node->GetINode()->SetMtl(m);
			}

			node->SetName(objectname);
			i->RedrawViews();

		}
		return 1;

	}
	return 0;
}

